<?php
/* Copyright (C) 2016 John Valance - All Rights Reserved.
   You may use, distribute and modify this code under the
   terms of the BSD license. See file LICENSE.txt in the
   top-most parent folder distributed with this source code.
   If missing, please write to: johnv@div1sys.com.
*/
require_once('../03-singleton/DB2Connection.php');
require_once('ToolkitService.php');
require_once 'Query_Builder.php';

class DBMapper {
    private $tableName;
    private $className;
    private $keyFields = array();
    private $metaData = array();
    private $hasAutoID = false;

    public function __construct(
        $className,
        $tableName,
        array $keyFields,
        $hasAutoID = false
        )
    {
        if (empty($tableName) ||
            empty($className) ||
            empty($keyFields) ||
            !is_array($keyFields) ||
            count($keyFields) == 0)
        {
            throw new Exception(
                'Class Name, Table Name, and at least one Key Field, are required to instantiate DBMapper object.');
        }

        $this->className = $className;
        $this->tableName = $tableName;
        $this->keyFields = $keyFields;
        $this->hasAutoID = $hasAutoID;

        $this->addMetaData( $tableName );
    }

    public function setHasAutoID( $hasAutoID ) {
        $this->hasAutoID = $hasAutoID;
    }
    public function getHasAutoID( ) {
        return $this->hasAutoID;
    }

    public function getTableName()
    { return $this->tableName; }

    public function getClassName()
    { return $this->className; }

    public function getKeyFields()
    { return $this->keyFields; }

    public function getMetaData()
    { return $this->metaData; }

    private function addMetaData($table) {

        try {
            $conn = DB2Connection::getInstance();
            $tk = ToolkitService::getInstance($conn, DB2_I5_NAMING_ON);
			//$tk->setOptions(array('stateless' => true));
            $objd = $tk->ClCommandWithOutput("RTVOBJD OBJ(*LIBL/$table) OBJTYPE(*FILE) RTNLIB(?)");
        } catch (Exception $e) {
            var_dump($e);
        }

        $schema = $objd['RTNLIB'];

        if ( empty($schema) ) {
            throw new Exception("Unable to find table metadata for $table. (FYI: Length of table name should not exceed 10; i.e. use system table name.)");
        }

        $table = strtoupper($table);
        $schema = strtoupper($schema);
        $sql = "select * from qsys2/syscolumns
                where table_schema = '$schema'
                and system_table_name = '$table' ";

        $stmt = db2_prepare($conn, $sql);
        if (!$stmt) throw new Exception("Select for metadata - Prepare failed! : SQL = $sql<br> ". db2_stmt_errormsg());

        if (!db2_execute($stmt))
            throw new Exception("db2_execute() for metadata failed! : SQL = $sql<br> ". db2_stmt_errormsg());

        while ($sysColumn = db2_fetch_assoc($stmt)) {
            // Add each column's metadata to the master metadata array
            $colName = $sysColumn['COLUMN_NAME'];
            $this->metaData[$colName] = $sysColumn;
        }
        //print_r($this->metaData);
    }

    /*
     * The mapper method
     */
    public function map( $data ) {
        // NOTE: assumes constructor has no required parameters
        $object = new $this->className;

        foreach ($this->metaData as $col => $colMeta) {
            $setterMethod = "set$col";
            $object->$setterMethod($data[$col]);
        }
        return $object;
    }

    /*
     * FINDER METHOD - SINGLE RECORD/OBJECT
     */
    public function find( array $keys )
    {
        $conn = DB2Connection::getInstance();

        $select_row_sql = "select * from {$this->tableName} ";
        $where = $this->getWhereClause($keys);
        $select_row_sql .= $where['whereString'];
        $bindVals = $where['whereValues'];

        $select_row_stmt = db2_prepare($conn, $select_row_sql);
        if (!$select_row_stmt) throw new Exception(
            "Select ROW prepare failed! : ". db2_stmt_errormsg()
        );

        if (!db2_execute($select_row_stmt, $bindVals))
            throw new Exception("db2_execute() failed! : ". db2_stmt_errormsg());

        $row = db2_fetch_assoc($select_row_stmt);

        if (!is_array($row)) {
            throw new Exception("{$this->tableName} record not found.");
        } else {
            return $this->map($row);
        }
    }

    /*
     * FINDER METHOD - MULTIPLE RECORDS/OBJECTS
     */
    public function find_multi( Query_Builder $qb )
    {
        $conn = DB2Connection::getInstance();

        $select_row_stmt = db2_prepare($conn, $qb->toString());
        if (!$select_row_stmt) throw new Exception(
            "Select ROW prepare failed! : ". db2_stmt_errormsg());

        if (!db2_execute($select_row_stmt, $qb->getWhereValues()))
            throw new Exception("db2_execute() failed! : ". db2_stmt_errormsg());

        $rows = array();
        while ($row = db2_fetch_assoc($select_row_stmt)) {
            $rows[] = $this->map($row);
        }

        return $rows;
    }

    public function find_row_count( Query_Builder $qb )
    {
        $conn = DB2Connection::getInstance();

        $select_row_stmt = db2_prepare($conn, $qb->toStringRowCount());
        if (!$select_row_stmt) throw new Exception(
            "Select ROW prepare failed! : ". db2_stmt_errormsg()
        );

        if (!db2_execute($select_row_stmt, $qb->getWhereValues()))
            throw new Exception("db2_execute() failed! : ". db2_stmt_errormsg());

        $row = db2_fetch_assoc($select_row_stmt);
        return (int) $row['ROW_COUNT'];
    }

    /*
     * Database updating methods
     */

    public function insert( $object )
    {
        $conn = DB2Connection::getInstance();

        $meta = $this->metaData;
        if ($this->hasAutoID) {
            // Remove any auto-generated key fields
            foreach ($this->keyFields as $keyField) {
                unset($meta[$keyField]);
            }
        }

        $cols = array_keys($meta);
        $insert_columns = implode(', ', $cols);

        $quests = array_fill(1, count($meta), '?');
        $quests2 = array_values($quests);
        $insert_quests = implode(', ', $quests2);

        $insert_sql = "insert into {$this->tableName} ($insert_columns) values ($insert_quests)";

        foreach ($meta as $col => $colMeta) {
            $getterMethod = "get{$col}";
            $bindVals[] = $object->$getterMethod();
        }

        $insert_stmt = db2_prepare($conn, $insert_sql);
        if (!$insert_stmt) throw new Exception("Prepare INSERT failed! : $insert_sql <br>". db2_stmt_errormsg());

        if (!db2_execute($insert_stmt, $bindVals))
            throw new Exception("db2_execute() failed! : ". db2_stmt_errormsg());

        if ($this->hasAutoID) {
            $autoID = db2_last_insert_id($conn);
            $setterMethod = "set{$this->keyFields[0]}";
            $object->$setterMethod($autoID);
            return $autoID;
        }

        return true;
    }

    public function update( $object )
    {
        $conn = DB2Connection::getInstance();

        $update_sql = "update {$this->tableName} set ";
        $separator = '';

        foreach ($this->metaData as $upd_col => $col_meta) {
            // Only add non key fields to update list
            if (!in_array($upd_col, $this->keyFields)) {
                // See if getter method is callable first
                $getterMethod = "get{$upd_col}";
                if ( is_callable( array($object, $getterMethod) ) ) {
                    $update_sql .= "$separator $upd_col = ?";
                    $separator = ', '; // field name separator
                    $bindVals[] = $object->$getterMethod();
                }
            }
        }

        $where = $this->getWhereClause($object);
        $update_sql .= $where['whereString'];
        $bindVals = array_merge($bindVals, $where['whereValues']);

//         echo '<p>SQL = ';
//         var_dump($update_sql);
//         var_dump($bindVals);
//         echo '</p>';

        $update_stmt = db2_prepare($conn, $update_sql);
        if (!$update_stmt) throw new Exception("Prepare UPDATE failed! : ". db2_stmt_errormsg());

        if (!db2_execute($update_stmt, $bindVals))
            throw new Exception("db2_execute() failed! : ". db2_stmt_errormsg());

        return true;
    }

    public function delete( $object ) {
        $conn = DB2Connection::getInstance();

        $delete_row_sql = "delete from {$this->tableName}  ";

        $where = $this->getWhereClause($object);
        $delete_row_sql .= $where['whereString'];
        $bindVals = $where['whereValues'];

        $delete_row_stmt = db2_prepare($conn, $delete_row_sql);
        if (!$delete_row_stmt) throw new Exception("Delete ROW prepare failed! : ". db2_stmt_errormsg());

        if (!db2_execute($delete_row_stmt, $bindVals))
            throw new Exception("Delete db2_execute() failed! : ". db2_stmt_errormsg());

    }


    /*
     * Helper methods for contructing sql WHERE clause
     */
    private function getWhereClause( $object ) {

        // If array of keys/values passed, pass control to static getWhereClauseArray() method
        if (is_array($object)) {
            return DBMapper::getWhereClauseArray($object);
        }

        $whereClause = ' where ';
        $separator = '';

        foreach ($this->keyFields as $keyField ) {
            // Add key fields to where clause
            $whereClause .= "$separator $keyField = ?";
            $separator = ' AND ';
            $getterMethod = "get{$keyField}";
            $bindVals[] = $object->$getterMethod();
        }

        $returnValues = array(
            'whereString' => $whereClause,
            'whereValues' => $bindVals
        );

        //var_dump($returnValues);

        return $returnValues;
    }


    public static function getWhereClauseArray( $keyData ) {

        $whereClause = ' where ';

        $first_key = true;
        foreach ($keyData as $keyName => $keyValue ) {
            // Add key fields to where clause
            if ($first_key) {
                $whereClause .= "$keyName = ?";
                $first_key = false;
            } else {
                $whereClause .= " and $keyName = ?";
            }
            $bindVals[] = $keyValue;
        }

        $returnValues = array(
            'whereString' => $whereClause,
            'whereValues' => $bindVals
        );

        return $returnValues;
    }



}
